@@ -26,6 +26,19 @@ class Comparisons:
     @staticmethod
     def get_date_comparison(format_string):        
         return lambda a, b: datetime.strptime(a, format_string) < datetime.strptime(b, format_string)
+       
+operator_functions = {
+    Operators.equal                 :   lambda t, v, c: t == v,
+    Operators.not_equal             :   lambda t, v, c: t != v,
+    Operators.less_than             :   lambda t, v, c: c(t, v),
+    Operators.greater_than          :   lambda t, v, c: c(v, t),
+    Operators.less_than_or_equal    :   lambda t, v, c: not c(v, t),
+    Operators.greater_than_or_equal :   lambda t, v, c: not c(t, v),
+    Operators.inside                :   lambda t, v, c: t in [str(x) for x in v],
+    Operators._not                  :   lambda t, v, c: not operator_functions[list(v)[0]](t, v[list(v)[0]], c),
+    Operators._and                  :   lambda t, v, c: not (False in [operator_functions[list(d)[0]](t, d[list(d)[0]], c) for d in v]),
+    Operators._or                   :   lambda t, v, c: True in [operator_functions[list(d)[0]](t, d[list(d)[0]], c) for d in v],
+}
 
 class Dataset:
 
@@ -109,14 +122,14 @@ class Dataset:
         
         return self
 
-    def query(self, query_object=None):
-        if query_object == None:
+    def query(self, filter_object=None):
+        if filter_object == None:
             return self
-        if type(query_object) is not dict:
+        if type(filter_object) is not dict:
             return Dataset()
-        for field, operators in query_object.items():
+        for field, operators in filter_object.items():
             if type(operators) is not dict:
-                query_object[field] = {Operators.equal: operators}
+                filter_object[field] = {Operators.equal: operators}
 
         def double_binary_search(key, conditions):
 
@@ -183,16 +196,18 @@ class Dataset:
             return self.data[low_edge:high_edge]
         
 
-        result_data = copy.deepcopy(self.data)
+        result_data = []
 
-        if self.indexed_field in query_object.keys():
-            result_data = double_binary_search(self.fields.index(self.indexed_field), query_object[self.indexed_field])
+        if self.indexed_field in filter_object.keys():
+            result_data = double_binary_search(self.fields.index(self.indexed_field), filter_object[self.indexed_field])
+        else:
+            result_data = copy.deepcopy(self.data)
 
         deletions = []
 
         for i, row in enumerate(result_data):
 
-            for field, operations in query_object.items():
+            for field, operations in filter_object.items():
 
                 if not field in self.fields:
                     error_message(f"Dataset.query: field \'{field}\' does not exist, skipping")
@@ -204,27 +219,14 @@ class Dataset:
                     def get_comparator(t, v):
                         if not Operators.comparison in operations:
                             error_message(f"Dataset.query.get_comparator: comparison not specified for \'{field}\' filter, using default comparison")
-                            query_object[field]["comparison"] = Comparisons.default # so the message doesn't appear again
+                            filter_object[field]["comparison"] = Comparisons.default # so the message doesn't appear again
                             return Comparisons.default(t, v)
                         comparator = operations[Operators.comparison]
                         if type(comparator) is not types.FunctionType:
                             error_message(f"Dataset.query.get_comparator: comparison for \'{field}\' filter is not of type 'FunctionType', using default comparison instead")
-                            query_object[field]["comparison"] = Comparisons.default
+                            filter_object[field]["comparison"] = Comparisons.default
                             return Comparisons.default(t, v)
                         return comparator(t, v)
-                    
-                    operator_functions = {
-                        Operators.equal                 :   lambda t, v, c: t == v,
-                        Operators.not_equal             :   lambda t, v, c: t != v,
-                        Operators.less_than             :   lambda t, v, c: c(t, v),
-                        Operators.greater_than          :   lambda t, v, c: c(v, t),
-                        Operators.less_than_or_equal    :   lambda t, v, c: not c(v, t),
-                        Operators.greater_than_or_equal :   lambda t, v, c: not c(t, v),
-                        Operators.inside                :   lambda t, v, c: t in [str(x) for x in v],
-                        Operators._not                  :   lambda t, v, c: not operator_functions[list(v)[0]](t, v[list(v)[0]], c),
-                        Operators._and                  :   lambda t, v, c: not (False in [operator_functions[list(d)[0]](t, d[list(d)[0]], c) for d in v]),
-                        Operators._or                   :   lambda t, v, c: True in [operator_functions[list(d)[0]](t, d[list(d)[0]], c) for d in v],
-                    }
 
                     if operator == Operators.comparison:
                         continue
@@ -233,7 +235,7 @@ class Dataset:
                             deletions.append(i)
                             break
                     else:
-                        error_message("Dataset.query: operator \'{operator}\' does not exist, skipping")
+                        error_message(f"Dataset.query: operator \'{operator}\' does not exist, skipping")
 
                 if i in deletions:
                     break
@@ -246,8 +248,8 @@ class Dataset:
         result.fields = self.fields
         return result
 
-    def query_one(self, query_object=None):
-        dataset = self.query(query_object)
+    def query_one(self, filter_object=None):
+        dataset = self.query(filter_object)
         if len(dataset.data) == 0:
             return dataset
         else:
@@ -285,6 +287,21 @@ class Dataset:
 
         return dataset
 
+    def select_unique(self, field):
+        selection = self.select(field)
+
+        occurences = []
+        deletions = []
+        for i, row in enumerate(selection.data):
+            if row[0] in occurences:
+                deletions.append(i)
+            else:
+                occurences.append(row[0])
+        for i in reversed(range(len(deletions))):
+            del selection.data[deletions[i]]
+        
+        return selection
+
     def add_field(self, field, derivation=lambda r:""):
         for row in self.data:
             r = {f: row[self.fields.index(f)] for f in self.fields}
@@ -318,21 +335,20 @@ class Dataset:
 
         return self
 
-    def replace_derived(self, field_names, function):
+    def replace_derived(self, field_names, derivation):
         field_ids = self.get_field_ids(field_names)
 
         for r, row in enumerate(self.data):
             for i in field_ids:
-                row[i] = function(self.row_to_dict(row))
+                row[i] = derivation(self.row_to_dict(row))
 
         return self
 
-    def join(self, other_dataset, common_fields):
+    def join(self, other_dataset, common_fields, remove=True):
         if type(common_fields) is str:
             common_fields = [common_fields] * 2
         for field in other_dataset.fields:
             if field == common_fields[1]: continue
-
             def match(row):
                 matched_row = other_dataset.query_one({
                     common_fields[1]: row[common_fields[0]]
@@ -341,9 +357,11 @@ class Dataset:
                     return matched_row.to_dictionary()[field]
                 else:
                     return ""
-
             self.add_field(field, match)
-        return self.remove_fields(common_fields[0])
+
+        if remove:
+            self.remove_fields(common_fields[0])
+        return self
 
     def to_dictionary(self):
         if len(self.data) > 1:
